/* PyV4L Version 0.4.1
Copyright (c) 2002 Michael Dove <pythondeveloper@optushome.com.au>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Derived from fmio <bktr.c> by Vladimir Popov <jumbo@narod.ru> */
 
 

#include <fcntl.h>
#include <linux/videodev.h>
#include <sys/ioctl.h>
#include <Python.h>
#include <stdio.h>
#include <sys/mman.h>
#include <math.h>
#define WIDTH 320
#define HEIGHT 240

/* radioobject information */
typedef struct {
    PyObject_HEAD
    int		fd;
    float	fact;
} radioobject;

/* videoobject information */
typedef struct {
    PyObject_HEAD
    int		fd;
    float	fact;
    char	*map;
    struct video_mbuf mbuf;
    struct video_mmap vm;
    float	fps;
    int		frame_count;
    clock_t	cstart;
    clock_t	clast;
} videoobject;


/* v4lobject information */
typedef struct {
    PyObject_HEAD
    int		fd;
    float	fact;
} v4lobject;

staticforward PyTypeObject V4LType;
staticforward PyTypeObject RadioType;
staticforward PyTypeObject VideoType;

/* Some Doc Strings */
static char radio_doc[] =
"v4l.radio(filedescriptor) -> radio object\n"
"\n"
"Control a v4l radio tuner.\n"
"Open a radio device file (eg. /dev/radio), pass device file name to constructor.";

static char video_doc[] =
"v4l.video(filedescriptor) -> video object\n"
"\n"
"Control a v4l video device.\n"
"Open a video device file (eg. /dev/video), pass device file name to constructor.";

static char v4l_doc[] =
"v4l.V4L(filedescriptor) -> generic V4L object\n"
"\n"
"Control a v4l video device.\n"
"Open a video device file (eg. /dev/video), pass device file name to constructor.\n";

static char getfreq_doc[] =
"getFrequency() -> int frequency (Khz)\n"
"\n"
"Gets the current frequency of device.";

static char getmode_doc[] =
"getMode() -> int mode\n"
"\n"
"Returns audio mode.";

static char getvol_doc[] =
"getVolume() -> int volume\n"
"\n"
"Returns volume level (0-9).";

static char setfreq_doc[] =
"setFrequency(integer) -> None\n"
"\n"
"Set Frequency (Khz)";

static char setvol_doc[] =
"setVolume(volumelevel) -> None\n"
"\n"
"Sets the volume level, not supported by all devices.\n";

static char mute_doc[] =
"mute() -> None\n"
"\n"
"Mutes device.";

static char setbass_doc[] =
"setBass(basslevel) -> None\n"
"\n"
"Sets the bass level (0-9), not supported by all devices.";

static char settreble_doc[] =
"setTreble(treblelevel) -> None\n"
"\n"
"Sets the treble level (0-9), not supported by all devices.";

static char setmono_doc[] =
"setMono() -> None\n"
"\n"
"Sets audio mode to mono.";

static char setstereo_doc[] =
"setStereo() -> None\n"
"\n"
"Sets audio mode to stereo, not supported by all devices.";

static char getbuffer_doc[] =
"getBuffer() -> (int height, int width, int depth, int bytesperline)\n"
"\n"
"Returns buffer tuple.";

static char getcapabilities_doc[] =
"getCapabilities() -> (string name, int type,\n int channels, int audios,\n int maxwidth, int maxheight,\n int minwidth, int minheight)\n"
"\n"
"Returns the device capabilities.";

static char getchannel_doc[] =
"getChannel(channel) -> (int channel, string name, int tuners)"
"\n"
"Returns channel information.";

static char getoverlay_doc[] =
"getOverlay() -> (int x, int y, int width, int height, int chromakey, int flags, int clipcount)\n"
"\n"
"Return overlay information.";

static char getpicture_doc[] =
"getPicture() -> (int brightness, int hue, int colour, int contrast, int whiteness, int depth, int palette)\n"
"\n"
"Returns picture information, whiteness only for black & white palette types.";

static char settuner_doc[] =
"setTuner() -> None\n"
"\n"
"Set tuner mode.";

static char gettuner_doc[] =
"getTuner() -> (int tuner, string name, int rangelow, int rangehigh, int flags, int mode, int signal)\n"
"\n"
"Returns tuner information.";



static char getvbi_doc[] =
"getVbi() -> (int sampling_rate, int samples_per_line,\n\
    int sample_format, (int start1, int start2),\n\
    (int count1, int count2), int flags)\n"
"\n"
"Returns VBI information.";

static char setchannel_doc[] =
"setChannel(channel) -> None\n"
"\n"
"Set Channel.";

static char setoverlay_doc[] =
"setOverlay(x, y, width, height) -> None\n"
"\n"
"Set Overlay information.";

static char setpicture_doc[] =
"setPicture(bright, hue, colour, contrast, whiteness, depth, palette) -> None\n"
"\n"
"Set Picture information.\n";

static char setsync_doc[] =
"setSync(sync) -> None\n"
"\n"
"Set sync.";

static char startcapture_doc[] =
"startCapture() -> None\n"
"\n"
"Start Capture.";

static char stopcapture_doc[] =
"stopCapture() -> None\n"
"\n"
"Stop Capture.";

static char prequeueframes_doc[] =
"preQueueFrames() -> None\n"
"\n"
"Queue up frames before capture.";

static char queueframe_doc[] = 
"queueFrame() -> None\n"
"\n"
"Queue up a single frame.";

static char getimage_doc[] = 
"getImage() -> string image\n"
"\n"
"Get a frame from device.";

static char setupimage_doc[] =
"setupImage(width, height) -> None\n"
"\n"
"Set image capture parameters.";

    


/* Instance methods
 * 	radio.setFrequency()	- set tuner frequency
 * 	radio.getFrequency()	- get tuner frequency
 * 	radio.mute()		- mute tuner
 * 	radio.setVolume()	- set volume
 * 	radio.getVolume()	- get volume
 * 	radio.setMono()		- set mode to Mono
 * 	radio.setStereo()	- set mode to Stereo
 * 	radio.getMode()		- get mode
 * 	radio.setBass()		- set bass level - hardware must support
 * 	radio.setTreble()	- set treble level - hardware must support
 */


/* Exceptions */
static PyObject *V4lError;
static PyObject *AudioError; /* Errors anything to do with audio */
static PyObject *VideoError; /* Errors anything to do with video */

/* Functions */
static PyObject *
v4l_getCapabilities(v4lobject *self) {
     struct video_capability vc;
     PyObject *cap;
     if (ioctl(self->fd, VIDIOCGCAP, &vc) < 0) {
	 PyErr_SetString(V4lError, "Error retriving video capability.");
	 return NULL;
     }
     cap = Py_BuildValue("(siiiiiii)", 
	     vc.name, 
	     vc.type, 
	     vc.channels, 
	     vc.audios, 
	     vc.maxwidth,
	     vc.maxheight,
	     vc.minwidth,
	     vc.minheight
	     );
     //printf("VID_TYPE_CHROMAKEY: %d\n", VID_TYPE_CHROMAKEY);
     return cap;
}
     
static PyObject *
v4l_getChannel(v4lobject *self, PyObject *args) {
    int chan;
    if (!PyArg_ParseTuple(args, "i", &chan)) return NULL;
    struct video_channel vc;
    vc.channel = chan;
    PyObject *vidchan;
    if (ioctl(self->fd, VIDIOCGCHAN, &vc) < 0) {
	PyErr_SetString(V4lError, "Error retriving video channel.");
	return NULL;
    }
    vidchan = Py_BuildValue("(isi)",
	    vc.channel,
	    vc.name,
	    vc.tuners
	    );
    return vidchan;
}




static PyObject *
v4l_setChannel(v4lobject *self, PyObject *args) {
    int chan;
    struct video_channel vc;
    if (!PyArg_ParseTuple(args, "i", &chan)) return NULL;
    //printf("Setting Video Channel: %d\n", chan);
    vc.channel = chan;
    if (ioctl(self->fd, VIDIOCGCHAN, &vc) < 0) {
	PyErr_SetString(V4lError, "Error retriving video channel.");
	return NULL;
    }
   
    //strcpy(vc.name, "Television");
    //vc.tuners = 1;
    if (ioctl(self->fd, VIDIOCSCHAN, &vc) < 0) {
	PyErr_SetString(V4lError, "Error setting channel.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
v4l_getTuner(v4lobject *self) {
    struct video_tuner vt;
    PyObject *tuner;
    vt.tuner = 0; // first tuner?
    if (ioctl(self->fd, VIDIOCGTUNER, &vt) < 0) {
	PyErr_SetString(V4lError, "Error retriving tuner.");
	return NULL;
    }
    tuner = Py_BuildValue("(icllihh)",
	    vt.tuner,
	    vt.name,
	    vt.rangelow,
	    vt.rangehigh,
	    vt.flags,
	    vt.mode,
	    vt.signal
	    );
    return tuner;
}

static PyObject *
v4l_setTuner(v4lobject *self, PyObject *args) {
    int mode;
    if (!PyArg_ParseTuple(args, "i", &mode)) return NULL;
    struct video_tuner vt;
    vt.tuner = 0; //first tuner?
    if (ioctl(self->fd, VIDIOCGTUNER, &vt) < 0) {
	PyErr_SetString(V4lError, "Error retriving tuner.");
	return NULL;
    }

    vt.mode = mode;

    if (ioctl(self->fd, VIDIOCSTUNER, &vt) < 0) {
	PyErr_SetString(V4lError, "Error setting tuner mode.");
	return NULL;
    }
			    
    Py_INCREF(Py_None);
    return Py_None;
}     
   

static PyObject *
v4l_getPicture(v4lobject *self) {
    struct video_picture vp;
    PyObject *pic;
    if (ioctl(self->fd, VIDIOCGPICT, &vp) < 0) {
	PyErr_SetString(V4lError, "Error retriving picture.");
	return NULL;
    }
    pic = Py_BuildValue("(hhhhhhh)",
	    vp.brightness,
	    vp.hue,
	    vp.colour,
	    vp.contrast,
	    vp.whiteness,
	    vp.depth,
	    vp.palette
	    );
    return pic;
}

static PyObject *
v4l_setPicture(v4lobject *self, PyObject *args) {
    unsigned int bright, hue, colour, contrast, whiteness, depth, palette;
    if (!PyArg_ParseTuple(args, "iiiiiii",
		&bright,
		&hue,
		&colour,
		&contrast,
		&whiteness,
		&depth,
		&palette
		)) return NULL;
    struct video_picture vp;
    vp.brightness = bright;
    vp.hue = hue;
    vp.colour = colour;
    vp.contrast = contrast;
    vp.whiteness = whiteness;
    vp.depth = depth;
    vp.palette = palette;

    if (ioctl(self->fd, VIDIOCSPICT, &vp) < 0) {
	PyErr_SetString(V4lError, "Error setting picture.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
v4l_startCapture(v4lobject *self) {
    if (ioctl(self->fd, VIDIOCCAPTURE, 1) < 0) {
	PyErr_SetString(V4lError, "Error starting video capture.");
	return NULL;
    }
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject *
v4l_stopCapture(v4lobject *self) {
    if (ioctl(self->fd, VIDIOCCAPTURE, 0) < 0) {
	PyErr_SetString(V4lError, "Error stopping video capture.");
	return NULL;
    }
   Py_INCREF(Py_None);
   return Py_None;
}

static PyObject *
v4l_getOverlay(v4lobject *self) {
    struct video_window vw;
    PyObject *overlay;
    if (ioctl(self->fd, VIDIOCGWIN, &vw) < 0) {
	PyErr_SetString(V4lError, "Error retriving overlay window.");
	return NULL;
    }
    overlay = Py_BuildValue("(iiiiiii)",
	    vw.x,
	    vw.y,
	    vw.width,
	    vw.height,
	    vw.chromakey,
	    vw.flags,
	    vw.clipcount
	    );
    return overlay;
}

static PyObject *
v4l_setOverlay(v4lobject *self, PyObject *args) {
    int x, y, width, height;
    if (!PyArg_ParseTuple(args, "iiii", &x, &y, &width, &height))
	return Py_BuildValue("i", 1);
    struct video_window vw;
    vw.x = x;
    vw.y = y;
    vw.width = width;
    vw.height = height;
    vw.flags = 0;
    vw.chromakey = 0;
    vw.clips = NULL;
    vw.clipcount = 0;

    if (ioctl(self->fd, VIDIOCSWIN, &vw) < 0) {
	PyErr_SetString(V4lError, "Error setting overlay window.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

//FIXME
static PyObject *
v4l_getBuffer(v4lobject *self) {
    struct video_buffer vb;
    PyObject *buffer;
    if (ioctl(self->fd, VIDIOCGFBUF, &vb) < 0) {
	PyErr_SetString(V4lError, "Error retriving video buffer.");
	return NULL;
    }
    buffer = Py_BuildValue("iiii",
	    vb.height,
	    vb.width,
	    vb.depth,
	    vb.bytesperline
	    );
    return buffer;
}

static PyObject *
v4l_setSync(v4lobject *self, PyObject *args) {
    int sync;
    if (!PyArg_ParseTuple(args, "i", &sync))
	return Py_BuildValue("i", 1);
    if (ioctl(self->fd, VIDIOCSYNC, &sync) < 0) {
	PyErr_SetString(V4lError, "Error syncing video.");
	return NULL;
    }
    Py_INCREF(Py_None); 
    return Py_None;
}

static PyObject *
v4l_getVbi(v4lobject *self) {
    struct vbi_format vf;
    PyObject * vbi;
    if (ioctl(self->fd, VIDIOCGVBIFMT, &vf) < 0) {
	PyErr_SetString(V4lError, "Error retriving VBI.");
	return NULL;
    }
    vbi = Py_BuildValue("(iii(ii)(ii)i)",
	    vf.sampling_rate,
	    vf.samples_per_line,
	    vf.sample_format,
	    vf.start[0],
	    vf.start[1],
	    vf.count[0],
	    vf.count[1],
	    vf.flags
	    );
    return vbi;
}
    
static PyObject *
v4l_setupImage(videoobject *self, PyObject *args) {
    unsigned int height;
    unsigned int width;
    unsigned int format = 0;
    if (!PyArg_ParseTuple(args, "ii|i", &width, &height, &format)) return NULL;
    self->vm.height = height;
    self->vm.width = width;
    if (format != 0)
    	self->vm.format = format;
    Py_INCREF(Py_None);
    return Py_None;
}


static int
v4l_setupCapture(videoobject *self) {
    if (ioctl(self->fd, VIDIOCGMBUF, &self->mbuf) < 0) return -1;
    //fprintf(stderr,"  mbuf: size=%d frames=%d\n", self->mbuf.size,self->mbuf.frames);
    self->map = mmap(0, self->mbuf.size, PROT_READ|PROT_WRITE, MAP_SHARED, self->fd, 0);
    if (self->map == NULL | self->map == MAP_FAILED) {
	PyErr_SetString(VideoError, "Error Memory Mapping device.");
    	return -1;
    }


    printf("Frame: %d\n", self->vm.frame);
    self->vm.frame = 0;

}

static PyObject *
v4l_preQueueFrames(videoobject *self) {
    unsigned int frame;
    if (self->map == NULL) if (v4l_setupCapture(self) < 0) return NULL;
    for (frame=0; frame < self->mbuf.frames; frame++) {
	self->vm.frame = frame;
     	if (ioctl(self->fd, VIDIOCMCAPTURE, &self->vm) < 0) {
	    PyErr_SetString(VideoError, "Error queuing frame.");
	    return NULL;
	}
    }
    Py_INCREF(Py_None);
    return Py_None;
}
	   
static PyObject *
v4l_queueFrame(videoobject *self) {
    unsigned short frame;
    if (self->map == NULL) if (v4l_setupCapture(self) < 0) return NULL;
    if (ioctl(self->fd, VIDIOCMCAPTURE, &self->vm) < 0) {
	PyErr_SetString(VideoError, "Error queuing frame.");
	return NULL;
    }
    frame = self->vm.frame + 1;
    if (frame >= self->mbuf.frames) frame = 0;
    return Py_BuildValue("h", frame); // next frame number
}
   
static PyObject *
v4l_getImage(videoobject *self, PyObject *args) {
    unsigned short frame = 0;
    if (!PyArg_ParseTuple(args, "|h", &frame)) return NULL;
    
    self->vm.frame = frame;
    if (ioctl(self->fd, VIDIOCSYNC,  &self->vm.frame) == -1) {
	//printf("Error: cannot sync\n");
	return NULL;
    }
    /*
    // FPS counter
    printf("FPS: %0.1f\n", self->fps);
    if ( ((self->cstart=clock())-self->clast) / CLOCKS_PER_SEC > 1) {
     	self->fps = (float) self->frame_count / ((float) (self->cstart-self->clast) / CLOCKS_PER_SEC);
        self->frame_count=0;
        self->clast=self->cstart;
    }
    self->frame_count++;
    */
      
	
    PyObject *result;	    
    result = Py_BuildValue("s#", self->map + self->mbuf.offsets[frame], self->vm.width * self->vm.height * 3);
    return (result);
} 
	        

/**** OLD FUNCTIONS ***/
static PyObject *
v4l_setFrequency(v4lobject *self, PyObject *args) {
    unsigned long basefreq;
    unsigned long freq;
    if (!PyArg_ParseTuple(args, "l", &basefreq)) return NULL;
    freq = basefreq * self->fact + 0.5 ;
    if (ioctl(self->fd, VIDIOCSFREQ, &freq) < 0) {
	PyErr_SetString(V4lError, "Error setting frequency.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
v4l_setRawFreq(v4lobject *self, PyObject *args) {
    unsigned long freq;
    if (!PyArg_ParseTuple(args, "l", &freq)) return NULL;
    if (ioctl(self->fd, VIDIOCSFREQ, &freq) < 0) {
	PyErr_SetString(V4lError, "Error setting frequency.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
v4l_getFrequency(v4lobject *self) {
    unsigned long freq;
    unsigned long basefreq;
    if (ioctl(self->fd, VIDIOCGFREQ, &freq) < 0) {
	PyErr_SetString(V4lError, "Error retriving frequency.");
	return NULL;
    }
    basefreq = rint(freq / self->fact);
    return Py_BuildValue("l", basefreq);
}

static PyObject *
v4l_mute(radioobject *self, PyObject *args) {
    struct video_audio va;
    if(!PyArg_ParseTuple(args, "")) return NULL;
    va.audio = 0;
    va.flags = VIDEO_AUDIO_MUTE;
    va.volume = 0;

    if (ioctl(self->fd, VIDIOCSAUDIO, &va) < 0) {
	PyErr_SetString(V4lError, "Error muting audio.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
v4l_setVolume(v4lobject *self, PyObject *args) {
    int vol;
    struct video_audio va;
    if(!PyArg_ParseTuple(args, "i", &vol)) return NULL;
    if (vol > 10) vol = 10;
    if (vol < 0) vol = 0;
    va.flags = VIDEO_AUDIO_VOLUME;
    va.audio = 0;
    va.volume = vol * (65535/10);

    if (ioctl(self->fd, VIDIOCSAUDIO, &va) < 0) {
	PyErr_SetString(V4lError, "Error setting volume.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
v4l_getVolume(v4lobject *self) {
    struct video_audio va;
    va.audio = 0;
    if (ioctl(self->fd, VIDIOCGAUDIO, &va) < 0) {
	PyErr_SetString(V4lError, "Error retriving volume.");
	return NULL;
    }
    return Py_BuildValue("i", va.volume * 10/65535);
}

static PyObject *
v4l_setMono(v4lobject *self, PyObject *args) {
    if(!PyArg_ParseTuple(args, "")) return NULL;
    struct video_audio va;
    va.audio = 0;
    va.mode = VIDEO_SOUND_MONO;

    if (ioctl(self->fd, VIDIOCSAUDIO, &va) < 0) {
	PyErr_SetString(V4lError, "Error setting audio to mono.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

	
static PyObject *
v4l_setStereo(v4lobject *self, PyObject *args) {
    if(!PyArg_ParseTuple(args, "")) return NULL;
    struct video_audio va;
    va.audio = 0;
    va.mode = VIDEO_SOUND_STEREO;

    if (ioctl(self->fd, VIDIOCSAUDIO, &va) < 0) {
	PyErr_SetString(V4lError, "Error setting audio to stereo.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
v4l_getMode(v4lobject *self) {
    struct video_audio va;
    va.audio = 0;
    if (ioctl(self->fd, VIDIOCGAUDIO, &va) > 0) {
	PyErr_SetString(V4lError, "Error retriving audio mode.");
	return NULL;
    }
    return Py_BuildValue("i", va.mode);
}


static PyObject *
v4l_setBass(v4lobject *self, PyObject *args) {
    int level;
    if(!PyArg_ParseTuple(args, "i", &level)) return NULL;
    struct video_audio va;
    if (level > 10) level = 10;
    if (level < 0 ) level = 0;
    va.flags = VIDEO_AUDIO_BASS;
    va.audio = 0;
    va.bass = level * (65535/10);

    if (ioctl(self->fd, VIDIOCSAUDIO, &va) > 0) {
	PyErr_SetString(V4lError, "Error setting bass level.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
v4l_setTreble(v4lobject *self, PyObject *args) {
    int level;
    if (!PyArg_ParseTuple(args, "i", &level)) return NULL;
    struct video_audio va;
    if (level > 10) level = 10;
    if (level < 0) level = 0;
    va.flags = VIDEO_AUDIO_TREBLE;
    va.audio = 0;
    va.treble = level * (65535/10);

    if (ioctl(self->fd, VIDIOCSAUDIO, &va) > 0) {
	PyErr_SetString(V4lError, "Error setting treble level.");
	return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}


float get_freq_fact(int fd) {
    struct video_tuner t;
    t.tuner = 0; // first tuner?
    if (ioctl(fd, VIDIOCGTUNER, &t) == -1 || (t.flags & VIDEO_TUNER_LOW) == 0)
	return .016;
    return 160.0;
}

/* Instance methods table. Used by v4l_getattr() */
static PyMethodDef v4l_methods[] = {
    {"setFrequency", (PyCFunction) v4l_setFrequency, METH_VARARGS, setfreq_doc},
    {"setRawFrequency", (PyCFunction) v4l_setRawFreq, METH_VARARGS, NULL},
    {"getFrequency", (PyCFunction) v4l_getFrequency, METH_NOARGS, getfreq_doc},
    {"mute", (PyCFunction) v4l_mute, METH_VARARGS, mute_doc},
    {"setVolume", (PyCFunction) v4l_setVolume, METH_VARARGS, setvol_doc},
    {"getVolume", (PyCFunction) v4l_getVolume, METH_NOARGS, getvol_doc},
    {"setMono", (PyCFunction) v4l_setMono, METH_VARARGS, setmono_doc},
    {"setStereo", (PyCFunction) v4l_setStereo, METH_VARARGS, setstereo_doc},
    {"getMode", (PyCFunction) v4l_getMode, METH_NOARGS, getmode_doc},
    {"setBass", (PyCFunction) v4l_setBass, METH_VARARGS, setbass_doc},
    {"setTreble", (PyCFunction) v4l_setTreble, METH_VARARGS, settreble_doc},
    { NULL, NULL }
};

static PyMethodDef radio_methods[] = {
    {"setFrequency", (PyCFunction) v4l_setFrequency, METH_VARARGS, setfreq_doc},
    {"getFrequency", (PyCFunction) v4l_getFrequency, METH_NOARGS, getfreq_doc},
    {"mute", (PyCFunction) v4l_mute, METH_VARARGS, mute_doc},
    {"setVolume", (PyCFunction) v4l_setVolume, METH_VARARGS, setvol_doc},
    {"getVolume", (PyCFunction) v4l_getVolume, METH_NOARGS, getvol_doc},
    {"setMono", (PyCFunction) v4l_setMono, METH_VARARGS, setmono_doc},
    {"setStereo", (PyCFunction) v4l_setStereo, METH_VARARGS, setstereo_doc},
    {"getMode", (PyCFunction) v4l_getMode, METH_NOARGS, getmode_doc},
    {"setBass", (PyCFunction) v4l_setBass, METH_VARARGS, setbass_doc},
    {"setTreble", (PyCFunction) v4l_setTreble, METH_VARARGS, settreble_doc},
    { NULL, NULL }
};

static PyMethodDef video_methods[] = {
    {"getCapabilities", (PyCFunction) v4l_getCapabilities, METH_VARARGS, setfreq_doc},
    {"getChannel", (PyCFunction) v4l_getChannel, METH_VARARGS, getchannel_doc},
    {"setChannel", (PyCFunction) v4l_setChannel, METH_VARARGS, setchannel_doc},
    {"setFrequency", (PyCFunction) v4l_setFrequency, METH_VARARGS, setfreq_doc},
    {"getFrequency", (PyCFunction) v4l_getFrequency, METH_NOARGS, getfreq_doc},
    {"setupImage", (PyCFunction) v4l_setupImage, METH_VARARGS, setupimage_doc},
    {"getTuner", (PyCFunction) v4l_getTuner, METH_NOARGS, gettuner_doc},
    {"setTuner", (PyCFunction) v4l_setTuner, METH_VARARGS, settuner_doc},
    {"getPicture", (PyCFunction) v4l_getPicture, METH_NOARGS, getpicture_doc},
    {"setPicture", (PyCFunction) v4l_setPicture, METH_VARARGS, setpicture_doc},
    {"setVolume", (PyCFunction) v4l_setVolume, METH_VARARGS, setvol_doc},
    {"getVolume", (PyCFunction) v4l_getVolume, METH_NOARGS, getvol_doc},
    {"mute", (PyCFunction) v4l_mute, METH_VARARGS, mute_doc},
//    {"startCapture", (PyCFunction) v4l_startCapture, METH_NOARGS, startcapture_doc},
//    {"stopCapture", (PyCFunction) v4l_stopCapture, METH_NOARGS, stopcapture_doc},
    {"getOverlay", (PyCFunction) v4l_getOverlay, METH_NOARGS, getoverlay_doc},
    {"setOverlay", (PyCFunction) v4l_setOverlay, METH_VARARGS, setoverlay_doc},
    {"getBuffer", (PyCFunction) v4l_getBuffer, METH_NOARGS, getbuffer_doc},
    {"setSync", (PyCFunction) v4l_setSync, METH_VARARGS, setsync_doc},
    {"getVbi", (PyCFunction) v4l_getVbi, METH_NOARGS, getvbi_doc},
    {"getImage", (PyCFunction) v4l_getImage, METH_VARARGS, getimage_doc},
    {"queueFrame", (PyCFunction) v4l_queueFrame, METH_NOARGS, queueframe_doc},
    {"preQueueFrames", (PyCFunction) v4l_preQueueFrames, METH_NOARGS, prequeueframes_doc},
    { NULL, NULL }
};
   

/* Basic Operations */

static v4lobject *
new_v4l(int fd) {
    v4lobject *self;
    struct video_channel vc;
    vc.channel = 0;
    if (ioctl(fd, VIDIOCGCHAN, &vc) < 0) {
	PyErr_SetString(V4lError, "No V4L device found.");
	return NULL;
    }
    self = PyObject_NEW(v4lobject, &V4LType);
    if (self == NULL) return NULL;
    self->fd = fd;
    self->fact = get_freq_fact(fd);
    return self;
}

static radioobject *
new_radio(int fd) {
    radioobject *self;
    struct video_tuner t;
    t.tuner = 0; //first device?
    // check to see if v4l is working
    if (ioctl(fd, VIDIOCGTUNER, &t) < 0) {
	PyErr_SetString(V4lError, "No V4L radio device found.");
	return NULL;
    }
    self = PyObject_NEW(radioobject, &RadioType);
    if (self == NULL) return NULL;
    self->fd = fd;
    self->fact = get_freq_fact(fd);
    return self;
}

static videoobject *
new_video(int fd) {
    videoobject *self;
    struct video_channel vc;
    vc.channel = 0;
    if (ioctl(fd, VIDIOCGCHAN, &vc) < 0) {
	PyErr_SetString(V4lError, "No V4L video device found.");
	return NULL;

	
    }
    self = PyObject_NEW(videoobject, &VideoType);
    if (self == NULL) return NULL;
    self->fd = fd;
    self->fact = get_freq_fact(fd);
    self->map = NULL;
    // Setup some defaults
    self->vm.height = HEIGHT;
    self->vm.width = WIDTH;
    self->vm.format = VIDEO_PALETTE_RGB24;
    self->cstart = 0;
    self->clast = 0;
    self->frame_count = 0;
    return self;
}

/* Release object */
static void
v4l_dealloc(v4lobject *self) {
    close(self->fd);
    PyMem_DEL(self);
}

static void
v4l_dealloc_video(videoobject *self) {
    if (self->map != NULL) {
	if (munmap(self->map, self->vm.height * self->vm.width * 3) < 0)
	    NULL;
	    //printf("Error Unmapping\n");
    }
    v4l_dealloc((v4lobject *)self);
}

/* Get an attribute */
static PyObject *
v4l_getattr(v4lobject *self, char *name) {
    if (strcmp(name, "fact") == 0) {
	return Py_BuildValue("d", self->fact); /* self.fact */
    } else if (strcmp(name, "fd") == 0) {
	return Py_BuildValue("i", self->fd); /* self.fd */
    }
    /* Look for a method instead */
    return Py_FindMethod(v4l_methods, (PyObject *)self, name);
}

static PyObject *
v4l_radiogetattr(radioobject *self, char *name) {
    /* Look for a method instead */
    return Py_FindMethod(radio_methods, (PyObject *)self, name);
}

static PyObject *
v4l_videogetattr(videoobject *self, char *name) {
    /* Look for a method instead */
    return Py_FindMethod(video_methods, (PyObject *)self, name);
}

/* repr() function */
static PyObject *
v4l_repr(v4lobject *self) {
    char rbuffer[256];
    sprintf(rbuffer, "<%s, fd = %d, fact = %.2f at %x>", self->ob_type->tp_name, self->fd, self->fact, self);
    return PyString_FromString(rbuffer);
}

/* Type object for v4l objects */
static PyTypeObject V4LType = {
    PyObject_HEAD_INIT(&PyType_Type) /* Required initialization */
    0,
    "v4l.V4L",
    sizeof(v4lobject),
    0,

    /* Standard methods */
    (destructor)  v4l_dealloc,	/* tp+dealloc,	: refcount = 0 	*/
    (printfunc)	  0,		/* tp_print	: print x	*/
    (getattrfunc) v4l_getattr, 	/* tp_getattr	: x.attr	*/
    (setattrfunc) 0,		/* tp_setattr	: x.attr = v	*/
    (cmpfunc)	  0, 		/* tp_compare	: x > y		*/
    (reprfunc)	  v4l_repr,	/* tp_repr	: repr(x)	*/

    /* Type categories */
    0,				/* tp_as_number	: Number methods */
    0,				/* tp_as_sequence : Sequence methods */
    0,		/* tp_hash	: dict[x]	*/
    0,		/* tp_call	: x()		*/
    0 		/* tp_str	: str(x)	*/
};

/* Type object for radio objects */
static PyTypeObject RadioType = {
    PyObject_HEAD_INIT(&PyType_Type) /* Required initialization */
    0,
    "v4l.radio",
    sizeof(radioobject),
    0,

    /* Standard methods */
    (destructor)  v4l_dealloc,	/* tp+dealloc,	: refcount = 0 	*/
    (printfunc)	  0,		/* tp_print	: print x	*/
    (getattrfunc) v4l_radiogetattr, 	/* tp_getattr	: x.attr	*/
    (setattrfunc) 0,		/* tp_setattr	: x.attr = v	*/
    (cmpfunc)	  0, 		/* tp_compare	: x > y		*/
    (reprfunc)	  v4l_repr,	/* tp_repr	: repr(x)	*/

    /* Type categories */
    0,				/* tp_as_number	: Number methods */
    0,				/* tp_as_sequence : Sequence methods */
    0,		/* tp_hash	: dict[x]	*/
    0,		/* tp_call	: x()		*/
    0 		/* tp_str	: str(x)	*/
};

/* Type object for radio objects */
static PyTypeObject VideoType = {
    PyObject_HEAD_INIT(&PyType_Type) /* Required initialization */
    0,
    "v4l.video",
    sizeof(videoobject),
    0,

    /* Standard methods */
    (destructor)  v4l_dealloc_video,	/* tp+dealloc,	: refcount = 0 	*/
    (printfunc)	  0,		/* tp_print	: print x	*/
    (getattrfunc) v4l_videogetattr, 	/* tp_getattr	: x.attr	*/
    (setattrfunc) 0,		/* tp_setattr	: x.attr = v	*/
    (cmpfunc)	  0, 		/* tp_compare	: x > y		*/
    (reprfunc)	  v4l_repr,	/* tp_repr	: repr(x)	*/

    /* Type categories */
    0,				/* tp_as_number	: Number methods */
    0,				/* tp_as_sequence : Sequence methods */
    0,		/* tp_hash	: dict[x]	*/
    0,		/* tp_call	: x()		*/
    0, 		/* tp_str	: str(x)	*/
    0,		/* tp_getattro			*/
    0,		/* tp_setattro			*/
    0,		/* tp_as_buffer			*/
    0,		/* tp_flags     		*/
};

/* Module Level functions */

/* Create a new v4l object as V4L(filename) */

static PyObject *
v4lobject_new(PyObject *self, PyObject *args) {
    int fd;
    char *fname;
    if (!PyArg_ParseTuple(args, "s", &fname)) return NULL;
    if (-1 == (fd = open(fname, O_RDWR))) {
	PyErr_SetString(PyExc_IOError, "Unable to open Device.");
	return NULL;
    }
    return (PyObject *) new_v4l(fd);
}

/* Create a new radio object as radio(filename) */

static PyObject *
radioobject_new(PyObject *self, PyObject *args) {
    int fd;
    char *fname;
    if (!PyArg_ParseTuple(args, "s", &fname)) return NULL;
    if (-1 == (fd = open(fname, O_RDWR))) {
	PyErr_SetString(PyExc_IOError, "Unable to open Device.");
	return NULL;
    }
    return (PyObject *) new_radio(fd);
}

/* Create a new video object as video(filename) */

static PyObject *
videoobject_new(PyObject *self, PyObject *args) {
    int fd;
    char *fname;
    if (!PyArg_ParseTuple(args, "s", &fname)) return NULL;
    if (-1 == (fd = open(fname, O_RDWR))) {
	PyErr_SetString(PyExc_IOError, "Unable to open Device.");
	return NULL;
    }
    return (PyObject *) new_video(fd);
}

/* Module Methods Table */
static struct PyMethodDef v4ltype_methods[] = {
    { "V4L", v4lobject_new, METH_VARARGS, v4l_doc },
    { "radio", radioobject_new, METH_VARARGS, radio_doc },
    { "video", videoobject_new, METH_VARARGS, video_doc },
    { NULL, NULL}
};

/* Module Description doc string */
static char module_doc[] =
"Implementation for V4L (Video4Linux) operations.";  


/* Module initialization function */
DL_EXPORT(void)
initv4l(void) {
    PyObject *m, *d;

    /* bind methods to types */
    RadioType.tp_methods = radio_methods;
    VideoType.tp_methods = video_methods;
    V4LType.tp_methods = v4l_methods;

    
    m = Py_InitModule3("v4l", v4ltype_methods, module_doc);
    d = PyModule_GetDict(m);

    V4lError = PyErr_NewException("v4l.V4lError", NULL, NULL);
    AudioError = PyErr_NewException("v4l.AudioError", NULL, NULL);
    VideoError = PyErr_NewException("v4l.VideoError", NULL, NULL);
    PyDict_SetItemString(d, "V4lError", V4lError);
    PyDict_SetItemString(d, "AudioError", AudioError);
    PyDict_SetItemString(d, "VideoError", VideoError);

    
    /* Add classes to module dictionary */
    //PyDict_SetItemString(d, "V4L", (PyObject*) &V4LType);
    //PyDict_SetItemString(d, "video", (PyObject*) &VideoType);
    //PyDict_SetItemString(d, "radio", (PyObject*) &RadioType);


    /* Constants */
    PyDict_SetItemString(d, "VIDEO_VC_TUNER", Py_BuildValue("i", VIDEO_VC_TUNER));
    PyDict_SetItemString(d, "VIDEO_VC_AUDIO", Py_BuildValue("i", VIDEO_VC_AUDIO));
    PyDict_SetItemString(d, "VIDEO_TYPE_TV", Py_BuildValue("i", VIDEO_TYPE_TV));
    PyDict_SetItemString(d, "VIDEO_TYPE_CAMERA", Py_BuildValue("i", VIDEO_TYPE_CAMERA));
    PyDict_SetItemString(d, "VID_TYPE_CHROMAKEY", Py_BuildValue("i", VID_TYPE_CHROMAKEY));


    PyDict_SetItemString(d, "VIDEO_TUNER_PAL", Py_BuildValue("i", VIDEO_TUNER_PAL));
    PyDict_SetItemString(d, "VIDEO_TUNER_NTSC", Py_BuildValue("i", VIDEO_TUNER_NTSC));
    PyDict_SetItemString(d, "VIDEO_TUNER_SECAM", Py_BuildValue("i", VIDEO_TUNER_SECAM));
    PyDict_SetItemString(d, "VIDEO_TUNER_LOW", Py_BuildValue("i", VIDEO_TUNER_LOW));

    PyDict_SetItemString(d, "VIDEO_TUNER_NORM", Py_BuildValue("i", VIDEO_TUNER_NORM));
    PyDict_SetItemString(d, "VIDEO_TUNER_STEREO_ON", Py_BuildValue("i", VIDEO_TUNER_STEREO_ON));
    PyDict_SetItemString(d, "VIDEO_TUNER_RDS_ON", Py_BuildValue("i", VIDEO_TUNER_RDS_ON));
    PyDict_SetItemString(d, "VIDEO_TUNER_MBS_ON", Py_BuildValue("i", VIDEO_TUNER_MBS_ON));

    PyDict_SetItemString(d, "VIDEO_MODE_PAL", Py_BuildValue("i", VIDEO_MODE_PAL));
    PyDict_SetItemString(d, "VIDEO_MODE_NTSC", Py_BuildValue("i", VIDEO_MODE_NTSC));
    PyDict_SetItemString(d, "VIDEO_MODE_SECAM", Py_BuildValue("i", VIDEO_MODE_SECAM));
    PyDict_SetItemString(d, "VIDEO_MODE_AUTO", Py_BuildValue("i", VIDEO_MODE_AUTO));

    PyDict_SetItemString(d, "VIDEO_PALETTE_GREY", Py_BuildValue("i", VIDEO_PALETTE_GREY));
    PyDict_SetItemString(d, "VIDEO_PALETTE_HI240", Py_BuildValue("i", VIDEO_PALETTE_HI240));
    PyDict_SetItemString(d, "VIDEO_PALETTE_RGB565", Py_BuildValue("i", VIDEO_PALETTE_RGB565));
    PyDict_SetItemString(d, "VIDEO_PALETTE_RGB24", Py_BuildValue("i", VIDEO_PALETTE_RGB24));
    PyDict_SetItemString(d, "VIDEO_PALETTE_RGB32", Py_BuildValue("i", VIDEO_PALETTE_RGB32));
    PyDict_SetItemString(d, "VIDEO_PALETTE_RGB555", Py_BuildValue("i", VIDEO_PALETTE_RGB555));
    PyDict_SetItemString(d, "VIDEO_PALETTE_YUV422", Py_BuildValue("i", VIDEO_PALETTE_YUV422));
    PyDict_SetItemString(d, "VIDEO_PALETTE_YUYV", Py_BuildValue("i", VIDEO_PALETTE_YUYV));
    PyDict_SetItemString(d, "VIDEO_PALETTE_UYVY", Py_BuildValue("i", VIDEO_PALETTE_UYVY));
    PyDict_SetItemString(d, "VIDEO_PALETTE_YUV420", Py_BuildValue("i", VIDEO_PALETTE_YUV420));
    PyDict_SetItemString(d, "VIDEO_PALETTE_YUV411", Py_BuildValue("i", VIDEO_PALETTE_YUV411));
    PyDict_SetItemString(d, "VIDEO_PALETTE_RAW", Py_BuildValue("i", VIDEO_PALETTE_RAW));
    PyDict_SetItemString(d, "VIDEO_PALETTE_YUV422P", Py_BuildValue("i", VIDEO_PALETTE_YUV422P));
    PyDict_SetItemString(d, "VIDEO_PALETTE_YUV411P", Py_BuildValue("i", VIDEO_PALETTE_YUV411P));
    PyDict_SetItemString(d, "VIDEO_PALETTE_YUV420P", Py_BuildValue("i", VIDEO_PALETTE_YUV420P));
    PyDict_SetItemString(d, "VIDEO_PALETTE_YUV410P", Py_BuildValue("i", VIDEO_PALETTE_YUV410P));
    PyDict_SetItemString(d, "VIDEO_PALETTE_PLANAR", Py_BuildValue("i", VIDEO_PALETTE_PLANAR));
    PyDict_SetItemString(d, "VIDEO_PALETTE_COMPONENT", Py_BuildValue("i", VIDEO_PALETTE_COMPONENT));
    
    PyDict_SetItemString(d, "VIDEO_AUDIO_MUTE", Py_BuildValue("i", VIDEO_AUDIO_MUTE));
    PyDict_SetItemString(d, "VIDEO_AUDIO_MUTABLE", Py_BuildValue("i", VIDEO_AUDIO_MUTABLE));
    PyDict_SetItemString(d, "VIDEO_AUDIO_VOLUME", Py_BuildValue("i", VIDEO_AUDIO_VOLUME));
    PyDict_SetItemString(d, "VIDEO_AUDIO_BASS", Py_BuildValue("i", VIDEO_AUDIO_BASS));
    PyDict_SetItemString(d, "VIDEO_AUDIO_TREBLE", Py_BuildValue("i", VIDEO_AUDIO_TREBLE));
    PyDict_SetItemString(d, "VIDEO_AUDIO_BALANCE", Py_BuildValue("i", VIDEO_AUDIO_BALANCE));
    PyDict_SetItemString(d, "VIDEO_SOUND_MONO", Py_BuildValue("i", VIDEO_SOUND_MONO));
    PyDict_SetItemString(d, "VIDEO_SOUND_STEREO", Py_BuildValue("i", VIDEO_SOUND_STEREO));
    PyDict_SetItemString(d, "VIDEO_SOUND_LANG1", Py_BuildValue("i", VIDEO_SOUND_LANG1));
    PyDict_SetItemString(d, "VIDEO_SOUND_LANG2", Py_BuildValue("i", VIDEO_SOUND_LANG2));


}
	

